;; guile-charting
;; Copyright (C) 2007 Andy Wingo <wingo at pobox dot com>

;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU Lesser General Public License as
;; published by the Free Software Foundation; either version 2.1 of the
;; License, or (at your option) any later version.
;;                                                                  
;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
;; Lesser General Public License for more details.
;;                                                                  
;; You should have received a copy of the GNU Lesser General Public
;; License along with this program; if not, contact:
;;
;; Free Software Foundation, Inc.     Voice:  +1-617-542-5942
;; 51 Franklin Street, Fifth Floor    Fax:    +1-617-542-2652
;; Boston, MA  02110-1301,  USA       gnu@gnu.org

;;; Commentary:
;;
;;
;;; Code:

(define-module (charting)
  :use-module (cairo)
  :use-module (charting util)
  :use-module (charting draw)
  :export (make-bar-chart make-chart))

(define-macro (with-move-to cr x y . body)
  `(begin
     (cairo-move-to ,cr ,x ,y)
     ,@body))

(define (fold proc seed list)
  (if (null? list)
      seed
      (fold proc (proc (car list) seed) (cdr list))))

(define-with-kwargs (make-chart title
                                (chart-height 360)
                                (chart-width 360)
                                (font-family "Bitstream Vera Sans")
                                (line-width 1)
                                (title-text-height 16)
                                (axis-text-height 12)
                                x-axis-label
                                y-axis-label
                                (tick-size 5)
                                y-axis-ticks
                                x-axis-ticks
                                y-axis-tick-labels
                                x-axis-tick-labels
                                (chart-margin 5)
                                (margin 5)
                                (make-surface
                                 (lambda (x y)
                                   (cairo-image-surface-create 'argb32 x y))))
  "Make a chart.

@var{tick-lables} is an alist of label-value pairs, where the
value is given in chart height coordinates. The label can be #f.

This function makes the basic chart, setting up the basics like
the title, axes, etc. You probably don't want to call this unless
you are making a custom chart type.

This function returns a cairo context whose coordinate system has
been flipped so that the origin of the chart is (0, 0), with
positive in the northeast quadrant.
"
  (define-with-kwargs (measure-text text (family font-family)
                                    (size axis-text-height))
    (if text
        (let ((cr (cairo-create (cairo-image-surface-create 'argb32
                                                            100 100))))
          (cairo-select-font-face cr family 'normal 'normal)
          (cairo-set-font-size cr size)
          (cairo-text-extents:width (cairo-text-extents cr text)))
        0))
  (let* ((y-axis-text-width (if y-axis-tick-labels
                                (apply max (map measure-text
                                                (map car y-axis-tick-labels)))
                                0))
         (total-width (+ chart-width (if y-axis-ticks tick-size 0)
                         (if y-axis-label axis-text-height 0)
                         y-axis-text-width
                         chart-margin chart-margin margin margin))

         (total-height (+ chart-height (if title title-text-height 0)
                          (if x-axis-ticks tick-size 0)
                          (if x-axis-tick-labels axis-text-height 0)
                          (if x-axis-label axis-text-height 0)
                          chart-margin chart-margin margin margin))
         (surface (make-surface total-width total-height))
         (cr (cairo-create surface)))

    ;; Move to cartesian coordinates centered at graph origin.
    (cairo-translate cr (+ margin chart-margin
                           (if y-axis-ticks tick-size 0)
                           (if y-axis-tick-labels y-axis-text-width 0)
                           (if y-axis-label axis-text-height 0))
                     (+ margin chart-margin
                        (if title title-text-height 0) chart-height))
    (cairo-scale cr 1.0 -1.0)

    (cairo-set-line-width cr line-width)
    (cairo-select-font-face cr font-family 'normal 'normal)

    (draw-background cr)
    (with-move-to
     cr 0 0
     (draw-chart-area cr chart-width chart-height))
    (with-move-to
     cr (/ chart-width 2) (+ chart-height chart-margin)
     (draw-title cr title title-text-height))
    
    (if x-axis-ticks
        (with-move-to
         cr 0 0
         (draw-ticks cr x-axis-ticks tick-size #f)))
    (if y-axis-ticks
        (with-move-to
         cr 0 0
         (draw-ticks cr y-axis-ticks tick-size #t)))
    (if x-axis-tick-labels
        (with-move-to
         cr 0 0
         (draw-tick-labels cr x-axis-tick-labels tick-size #f
                           axis-text-height)))
    (if y-axis-tick-labels
        (with-move-to
         cr 0 0
         (draw-tick-labels cr y-axis-tick-labels tick-size #t
                           axis-text-height)))
    (if x-axis-label
        (with-move-to
         cr 0 (- (+ axis-text-height tick-size))
         (draw-axis-label cr x-axis-label axis-text-height chart-height #f)))
    (if y-axis-label
        (with-move-to
         cr (- 0 y-axis-text-width tick-size) 0
         (draw-axis-label cr y-axis-label axis-text-height chart-height #t)))

    cr))

(define (make-uniform-ticks min max step)
  (let lp ((pos min) (out '()))
    (if (> pos max)
        (reverse! out)
        (lp (+ pos step)
            (cons pos out)))))

(define-with-kwargs (make-bar-chart title
                                    data 
                                    write-to-png
                                    (bar-width 40)
                                    (group-spacing 40)
                                    (chart-height 360)
                                    (chart-params '())
                                    (legend-params '()))
  "Make a bar chart.

The format of @var{data} is defined as follows:

@table @var
@item data
 (@var{group}+)
@item group
 (@var{group-label} @var{bar}+)
@item group-label
A string, to be written to the X axis.
@item bar
 (@var{height} @var{bar-params}?)
@item height
The bar height, as a number.
@item bar-params
A property list suitable to passing to
@ref{charting draw draw-bar}.
@end table

This function returns the cairo surface. By default, make-chart
will create an image surface, but you may override this by
passing a @code{:make-surface} function in the
@var{chart-params}. In this way you can render charts to any
surface supported by Cairo, e.g. PS, PDF, SVG, GDK, etc.

The :write-to-png option will write the chart out to the PNG file
that you name.

An example invocation might look like:

@example
 (make-bar-chart \"Average Height at Iihenda JSS\"
                 '((\"Grade 9\" (150 \"Boys\") (140 \"Girls\"))
                   (\"Grade 10\" (150 \"Boys\")
                               (140 \"Girls\" (:y+-bracket 5 :y--bracket 4.5))))
                 :write-to-png \"/tmp/graph.png\")
@end example"
  (let* ((num-groups (length data))
         (max-group-bars (1- (apply max (map length data))))
         (chart-width (* num-groups (+ (* max-group-bars bar-width)
                                       group-spacing)))
         (max-height (* (apply max (map (lambda (x) 
                                          (apply max (map car (cdr x))))
                                        data))
                        5/4))
         (height-scale (/ chart-height max-height))
         (yticks-unscaled (make-uniform-ticks
                           0 max-height
                           ((lambda (x) (if (<= (/ max-height x) 5)
                                            (/ x 2)
                                            x))
                            (expt 10 (1- (round
                                          (log10 max-height)))))))
         (yticks (map (lambda (x) (* x height-scale)) yticks-unscaled))
         (ytick-labels (map (lambda (pos) (cons (format #f "~g" pos)
                                                (* pos height-scale)))
                            yticks-unscaled))
         (xticks (make-uniform-ticks
                  0 chart-width
                  (+ (* max-group-bars bar-width) group-spacing)))
         (xtick-labels (let ((step (+ (* max-group-bars bar-width)
                                      group-spacing)))
                         (map cons
                              (map car data)
                              (make-uniform-ticks
                               (/ step 2) chart-width step))))
         (series-list (reverse! 
                       (fold (lambda (group series-list)
                               (fold
                                (lambda (bar series-list)
                                  (if (not (member (cadr bar) series-list))
                                      (cons (cadr bar) series-list)
                                      series-list))
                                series-list
                                (cdr group)))
                             '()
                             data)))
         (cr (apply make-chart title chart-height chart-width
                    :y-axis-ticks yticks :y-axis-tick-labels ytick-labels
                    :x-axis-ticks xticks :x-axis-tick-labels xtick-labels
                    chart-params)))

    (for-each
     (lambda (group pos)
       (with-move-to
        cr pos 0
        (draw-bar-group cr (cdr group) bar-width height-scale)))
     data
     (map cdr xtick-labels))

    (with-move-to
     cr 5 (- chart-height 5)
     (apply draw-legend cr #t #t :series-list series-list legend-params))

    (if write-to-png
        (cairo-surface-write-to-png (cairo-get-target cr) write-to-png))
    (cairo-get-target cr)))
