;; guile-charting
;; Copyright (C) 2007 Andy Wingo <wingo at pobox dot com>

;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU Lesser General Public License as
;; published by the Free Software Foundation; either version 2.1 of the
;; License, or (at your option) any later version.
;;                                                                  
;; This program is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
;; Lesser General Public License for more details.
;;                                                                  
;; You should have received a copy of the GNU Lesser General Public
;; License along with this program; if not, contact:
;;
;; Free Software Foundation, Inc.     Voice:  +1-617-542-5942
;; 51 Franklin Street, Fifth Floor    Fax:    +1-617-542-2652
;; Boston, MA  02110-1301,  USA       gnu@gnu.org

;;; Commentary:
;;
;; 
;;
;;; Code:

(define-module (charting draw)
  :use-module (ice-9 format)
  :use-module (cairo)
  :use-module (charting util)
  :export (draw-legend
           draw-bar
           draw-axis-label
           draw-bar-group
           draw-decorator
           draw-ticks
           draw-tick-labels
           draw-title
           draw-chart-area
           draw-background
           ))

(define pi 3.141592653589793)

(define-macro (with-cairo cr . body)
  `(begin
     (cairo-save ,cr)
     (let ((res (begin ,@body)))
       (cairo-restore ,cr)
       res)))

(define-macro (with-rgb cr r g b . body)
  `(with-cairo ,cr 
     (cairo-set-source-rgb ,cr ,r ,g ,b)
     ,@body))

(define-macro (with-rgba cr r g b a . body)
  `(with-cairo ,cr 
     (cairo-set-source-rgba ,cr ,r ,g ,b ,a)
     ,@body))

(define-macro (with-line-width cr w . body)
  `(with-cairo ,cr 
     (cairo-set-line-width ,cr ,w)
     ,@body))

(define-macro (with-font-size cr s . body)
  `(with-cairo ,cr 
     (cairo-set-font-size ,cr ,s)
     ,@body))

(define-macro (with-y-inverted cr . body)
  `(with-cairo ,cr 
     (cairo-scale cr 1 -1)
     ,@body))

(define-macro (with-rotate cr degrees . body)
  `(with-cairo ,cr 
     (cairo-rotate ,cr (* 2 ,pi (/ ,degrees 360)))
     ,@body))

(define-macro (with-color cr color . body)
  `(with-cairo ,cr
     (apply cairo-set-source-rgb ,cr
            (map 
             (lambda (x) (/ x 255.0))
             (assoc-ref *palette* ,color)))
     ,@body))

(define-macro (with-color-for-series cr series . body)
  `(with-color ,cr (color-for-series ,series)
               ,@body))

(define-macro (with-current-point cr x y . body)
  `(call-with-values (lambda () (cairo-get-current-point ,cr))
     (lambda (,x ,y) ,@body)))

;; the tango palette, mixed up a bit
(define *palette*
  '(("Chameleon 1"   138 226  52)
    ("Sky Blue 1"    114 159 207)
    ("Plum 1"        173 127 168)
    ("Scarlet Red 1" 239  41  41)
    ("Chocolate 1"   233 185 110)
    ("Orange 1"      252 175  62)
    ("Aluminium 1"   238 238 236)
    ("Butter 1"      252 233  79)
    ("Chameleon 2"   115 210  22)
    ("Chameleon 3"    78 154   6)
    ("Orange 2"      245 121   0)
    ("Aluminium 4"   136 138 133)
    ("Sky Blue 2"     52 101 164)
    ("Chocolate 2"   193 125  17)
    ("Sky Blue 3"     32  74 135)
    ("Aluminium 5"    85  87  83)
    ("Chocolate 3"   143  89   2)
    ("Plum 2"        117  80 123)
    ("Plum 3"         92  53 102)
    ("Orange 3"      206  92   0)
    ("Scarlet Red 2" 204   0   0)
    ("Scarlet Red 3" 164   0   0)
    ("Aluminium 2"   211 215 207)
    ("Butter 2"      237 212   0)
    ("Butter 3"      196 160   0)
    ("Aluminium 3"   186 189 182)
    ("Aluminium 6"    46  52  54)))

(define *series-colors* '())
(define *available-colors* *palette*)
  
(define (color-for-series series)
  (or (assoc-ref *series-colors* series)
      (begin
        (set! *series-colors*
              (acons series (caar *available-colors*)
                     *series-colors*))
        (set! *available-colors*
              (cdr *available-colors*))
        (cdar *series-colors*))))
  
(define (rel-rectangle cr width height)
  (cairo-rel-line-to cr width 0)
  (cairo-rel-line-to cr 0 height)
  (cairo-rel-line-to cr (- width) 0)
  (cairo-rel-line-to cr 0 (- height)))

(define (filled-rectangle cr width height stroke fill)
  (rel-rectangle cr width height)
  (with-color cr fill
              (cairo-fill-preserve cr))
  (with-color cr stroke
              (cairo-stroke cr)))

(define (filled-rectangle/1 cr width height stroke fill)
  (let ((line-width (cairo-get-line-width cr)))
    (cairo-rel-move-to cr (/ line-width 2) (/ line-width 2))
    (filled-rectangle cr (- width line-width)
                      (- height line-width)
                      stroke fill)))

(define (show-text cr text justification)
  (with-cairo cr
              (let ((width (cairo-text-extents:width
                            (cairo-text-extents cr text))))
                (cairo-rel-move-to
                 cr
                 (case justification
                   ((left) 0)
                   ((right) (- width))
                   ((center) (- (/ width 2)))
                   (else
                    (error "unknown justification" justification)))
                 0)
                (cairo-scale cr 1 -1)
                (cairo-show-text cr text))))

(define-with-kwargs (draw-legend cr expand-right? expand-down? (text-height 12)
                                 (series-list '()))
  "Draw a legend.

@var{series-list} is expected to be a list of series names. The
@var{cr} is expected to be positioned at one of the corners of
the legend; @var{expand-right?} and @var{expand-down?} control which way
the legend will be rendered."
  (let* ((box-width text-height)
         (padding 5)
         (spacing (/ text-height 2))
         (text-width (apply max (map (lambda (x)
                                       (with-font-size
                                        cr text-height
                                        (cairo-text-width cr x)))
                                     series-list)))
         (total-width (+ (* 2 padding) box-width spacing text-width))
         (num-series (length series-list))
         (total-height (+ (* 2 padding) (* (- num-series 1) spacing)
                          (* num-series text-height))))
    (with-current-point
     cr basex basey
     (cairo-rel-move-to cr (if expand-right? 0 (- total-width))
                        (if expand-down? (- total-height) 0))
     (filled-rectangle/1 cr total-width total-height
                         "Aluminium 6" "Aluminium 1")
     (let lp ((x (+ basex (if expand-right?
                              padding
                              (- (- total-width padding)))))
              (y (+ basey (if expand-down?
                              (- padding)
                              (- total-height padding))))
              (series-list series-list))
       (cond
        ((not (null? series-list))
         (cairo-move-to cr x (- y box-width))
         (filled-rectangle/1 cr box-width box-width
                         "Aluminium 6"
                         (color-for-series (car series-list)))
         (cairo-move-to cr (+ x box-width spacing) (- y text-height))
         (with-font-size cr text-height
                         (show-text cr (car series-list) 'left))
         (lp x (- y text-height spacing) (cdr series-list))))))))

(define-with-kwargs (draw-decorator cr scale y+-bracket y--bracket
                                    y-bracket)
  "Draw a decorator.

A decorator is something drawn around a point, such as error
bars. This function currently supports drawing error bars in the
Y direction, which are specified individually as @var{y+-bracket}
and @var{y--bracket}."
  (define (make-bracket x y mx my len)
    (cairo-move-to cr x y)
    (cairo-rel-line-to cr (* mx len) (* my len))
    (cairo-rel-line-to cr (* my -5) (* mx 0))
    (cairo-rel-line-to cr (* my 10) (* mx 0))
    (with-color cr "Aluminium 6" (cairo-stroke cr)))
  (with-current-point
   cr x y
   (if (or y+-bracket y-bracket)
       (make-bracket x y 0 1 (* (or y+-bracket y-bracket) scale)))
   (if (or y+-bracket y-bracket)
       (make-bracket x y 0 -1 (* (or y--bracket y-bracket) scale)))))

(define-with-kwargs (draw-bar cr height scale bar-width series (decorator '()))
  "Draw a single bar.

@var{cr} is expected to have been placed at the lower left corner
of where the bar should be. @var{decorator} is a property list
that can be passed to @ref{charting draw draw-decorator}."
  (with-current-point
   cr x y
   (filled-rectangle/1 cr bar-width (* height scale)
                       "Aluminium 6"
                       (color-for-series series))
   (cairo-move-to cr (+ x (/ bar-width 2)) (+ y (* height scale)))
   (apply draw-decorator cr scale decorator)))

(define-with-kwargs (draw-bar-group cr data bar-width scale)
  "Draw a group of bars.

@var{data} is a property list suitable for passing to
@ref{charting draw draw-bar}. @var{cr} is expected to have been positioned
along the x axis in the center of where the bar group should be
displayed."
  (cairo-rel-move-to cr (- (/ (* (length data) bar-width) 2)) 0)
  (with-current-point
   cr basex basey
   (let lp ((data data) (n 0))
     (cond
      ((not (null? data))
       (cairo-move-to cr (+ basex (* n bar-width)) basey)
       (apply draw-bar cr (caar data) scale bar-width (cdar data))
       (lp (cdr data) (1+ n)))))))
    
(define-with-kwargs (draw-ticks cr ticks tick-size vertical?)
  "Draw ticks on an axis.

@var{ticks} is a list of positions in the current cairo
coordinate system."
  (for-each
   (lambda (pos)
     (cond
      (vertical?
       (cairo-move-to cr 0 pos)
       (cairo-rel-line-to cr (- (* 2/3 tick-size)) 0))
      (else
       (cairo-move-to cr pos 0)
       (cairo-rel-line-to cr 0 (- (* 2/3 tick-size))))))
   ticks)
  (cairo-stroke cr))

(define-with-kwargs (draw-tick-labels cr tick-labels tick-size vertical?
                                      text-height)
  "Draw tick labels on an axis.

@var{tick-labels} is an alist of label-position pairs, where
position is in the current cairo coordinate system, along one
axis."
  (for-each
   (lambda (pair)
     (cond
      (vertical?
       (cairo-move-to cr (- tick-size) (- (cdr pair) (/ text-height 2)))
       (with-font-size
        cr text-height
        (show-text cr (car pair) 'right)))
      (else
       (cairo-move-to cr (cdr pair) (- (+ text-height tick-size)))
       (with-font-size
        cr text-height
        (show-text cr (car pair) 'center)))))
   tick-labels))

(define-with-kwargs (draw-axis-label cr text text-height axis-length vertical?)
  "Draw an axis label.

The label will be drawn such that the current position of
@var{cr} is the closest corner of the label's bounding box."
  (cond
   (vertical?
    (with-rotate
     cr 90
     (with-font-size
      cr text-height
      (cairo-rel-move-to cr (/ axis-length 2) 0)
      (show-text cr text 'center))))
   (else
    (with-font-size
     cr text-height
     (cairo-rel-move-to cr (/ axis-length 2) (- text-height))
     (show-text cr text 'center)))))
                                 
(define-with-kwargs (draw-title cr text font-size)
  "Draw a title.

@var{cr} is expected to have been positioned at the lower
boundary of where the title should be written, in the center."
  (with-font-size
   cr font-size
   (with-color cr "Aluminium 6" (show-text cr text 'center))))

(define-with-kwargs (draw-chart-area cr width height)
  "Draw the actual box for the chart background.

@var{cr} is expected to have been positioned at the origin."
  (with-line-width cr (* 2 (cairo-get-line-width cr))
                   (filled-rectangle cr width height
                                     "Aluminium 6"
                                     "Aluminium 1")))

(define-with-kwargs (draw-background cr)
  "Draw the background."
  (with-rgb cr 1 1 1
            (cairo-paint cr)))
